import yaml
from pysimmmulator.param_handlers import (
    basic_parameters,
    baseline_parameters,
    ad_spend_parameters,
    media_parameters,
    cvr_parameters,
    adstock_parameters,
    output_parameters,
)


def load_config(config_path: str) -> dict:
    with open(config_path) as f:
        cfg = yaml.load(f, Loader=yaml.FullLoader)
    return cfg


def define_basic_params(
    years,
    channels_clicks,
    channels_impressions,
    frequency_of_campaigns,
    start_date,
    true_cvr,
    revenue_per_conv,
):
    "Takes in requirements for basic_params and loads with dataclass for validation as precursor"
    my_basic_params = basic_parameters(
        years=years,
        channels_clicks=channels_clicks,
        channels_impressions=channels_impressions,
        frequency_of_campaigns=frequency_of_campaigns,
        start_date=start_date,
        true_cvr=true_cvr,
        revenue_per_conv=revenue_per_conv,
    )

    return my_basic_params

def validate_config(config_path: str, return_individual_results: bool = False):
    cfg = load_config(config_path=config_path)
    results = {}
    overall = True
    try:
        define_basic_params(**cfg["basic_params"])
        results.update({"basic_params":True})
    except Exception as e:
        results.update({"basic_params":e})
        overall = False
    try:
        my_basic_params = define_basic_params(**cfg["basic_params"])
        baseline_parameters(basic_params=my_basic_params, **cfg["baseline_params"])
        results.update({"baseline_params":True})
    except Exception as e:
        results.update({"baseline_params":e})
        overall = False
    
    matched_validation = {
        ad_spend_parameters:"ad_spend_params",
        media_parameters:"media_params",
        cvr_parameters:"cvr_params",
        adstock_parameters:"adstock_params",
        output_parameters:"output_params"
    }
    for handler, conf_name in matched_validation.items():
        try:
            handler(**cfg[conf_name])
            results.update({conf_name: True})
        except Exception as e:
            results.update({conf_name: e})
            overall = False

    if return_individual_results:
        return results
    else:
        return overall


import pysimmmulator as pysimmm

def test_viz_clicks_daily():
    cfg = pysimmm.load_parameters.load_config(config_path="example_config.yaml")
    sim = pysimmm.simmm()
    sim.run_with_config(config=cfg)
    sim.plot_clicks(agg='daily')

def test_viz_clicks_weekly():
    cfg = pysimmm.load_parameters.load_config(config_path="example_config.yaml")
    sim = pysimmm.simmm()
    sim.run_with_config(config=cfg)
    sim.plot_clicks(agg='weekly')

def test_viz_clicks_monthly():
    cfg = pysimmm.load_parameters.load_config(config_path="example_config.yaml")
    sim = pysimmm.simmm()
    sim.run_with_config(config=cfg)
    sim.plot_clicks(agg='monthly')

def test_viz_clicks_yearly():
    cfg = pysimmm.load_parameters.load_config(config_path="example_config.yaml")
    sim = pysimmm.simmm()
    sim.run_with_config(config=cfg)
    sim.plot_clicks(agg='yearly')

def test_viz_impressions_daily():
    cfg = pysimmm.load_parameters.load_config(config_path="example_config.yaml")
    sim = pysimmm.simmm()
    sim.run_with_config(config=cfg)
    sim.plot_impressions(agg='daily')

def test_viz_spend_daily():
    cfg = pysimmm.load_parameters.load_config(config_path="example_config.yaml")
    sim = pysimmm.simmm()
    sim.run_with_config(config=cfg)
    sim.plot_spend(agg='daily')
    



from pysimmmulator import load_parameters


def test_load_cfg():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    assert type(cfg) == dict


def test_cfg_base_keys():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    assert "basic_params" in cfg.keys()


def test_media_cfg_check():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    media_params = load_parameters.media_parameters(**cfg["media_params"])
    media_params.check(basic_params=my_basic_params)
    assert True == True


def test_cvr_cfg_check():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    cvr_params = load_parameters.cvr_parameters(**cfg["cvr_params"])
    cvr_params.check(basic_params=my_basic_params)
    assert True == True


def test_adstock_cfg_check():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    adstock_params = load_parameters.adstock_parameters(**cfg["adstock_params"])
    adstock_params.check(basic_params=my_basic_params)
    assert True == True

def test_validate_config():
    overall = load_parameters.validate_config(config_path="./example_config.yaml")
    assert overall == True


from pysimmmulator import load_parameters, simmm


def test_initiate_sim():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    simmm(my_basic_params)


def test_step1_baseline():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])


def test_step2_adspend():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])


def test_step3_media():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])
    sim.simulate_media(**cfg["media_params"])


def test_step4_cvr():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])
    sim.simulate_media(**cfg["media_params"])
    sim.simulate_cvr(**cfg["cvr_params"])


def tests_step5_adstock():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])
    sim.simulate_media(**cfg["media_params"])
    sim.simulate_cvr(**cfg["cvr_params"])
    sim.simulate_decay_returns(**cfg["adstock_params"])


def tests_step6_conversions():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])
    sim.simulate_media(**cfg["media_params"])
    sim.simulate_cvr(**cfg["cvr_params"])
    sim.simulate_decay_returns(**cfg["adstock_params"])
    sim.calculate_conversions()


def tests_step7_consolidatedataframe():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])
    sim.simulate_media(**cfg["media_params"])
    sim.simulate_cvr(**cfg["cvr_params"])
    sim.simulate_decay_returns(**cfg["adstock_params"])
    sim.calculate_conversions()
    sim.consolidate_dataframe()


def tests_step8_roi():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])
    sim.simulate_media(**cfg["media_params"])
    sim.simulate_cvr(**cfg["cvr_params"])
    sim.simulate_decay_returns(**cfg["adstock_params"])
    sim.calculate_conversions()
    sim.consolidate_dataframe()
    sim.calculate_channel_roi()


def tests_step9_consolidatedataframe():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    my_basic_params = load_parameters.define_basic_params(**cfg["basic_params"])
    sim = simmm(my_basic_params)
    sim.simulate_baseline(**cfg["baseline_params"])
    sim.simulate_ad_spend(**cfg["ad_spend_params"])
    sim.simulate_media(**cfg["media_params"])
    sim.simulate_cvr(**cfg["cvr_params"])
    sim.simulate_decay_returns(**cfg["adstock_params"])
    sim.calculate_conversions()
    sim.consolidate_dataframe()
    sim.finalize_output(**cfg["output_params"])


def test_run_with_config():
    cfg = load_parameters.load_config(config_path="example_config.yaml")
    sim = simmm()
    sim.run_with_config(config=cfg)


def test_run_with_config_weekly():
    cfg = load_parameters.load_config(config_path="example_config.yaml")
    cfg["output_params"]["aggregation_level"] = "weekly"
    sim = simmm()
    sim.run_with_config(config=cfg)


def test_run_with_high_frequency():
    cfg = load_parameters.load_config(config_path="example_config.yaml")
    cfg["basic_params"]["frequency_of_campaigns"] = 50
    sim = simmm()
    sim.run_with_config(config=cfg)
    assert sim.final_df.dropna().shape[0] > sim.final_df.shape[0] - 5



__title__ = "PySiMMM"
__author__ = "RyanAugust"
__license__ = "MIT"
__copyright__ = "Copyright 2024"

import os

with open(os.path.join(os.path.dirname(__file__), "VERSION")) as version_file:
    __version__ = version_file.read().strip()


from .simulate import simmm, multisimmm
from .load_parameters import load_config, define_basic_params



from pysimmmulator import load_config, simmm
import pandas as pd
import argparse
import os


def run_with_config(config_path, output_path):
    """Initiates `simmmulator` and executes `simmmmulator.run_with_config` against the passed config filepath

    Args:
        config_path (str): path to a valid config file, see example_config.yaml as example of `simmmulator` expected config format
    """
    cfg = load_config(config_path)
    sim = simmm()
    (mmm_input_df, channel_roi) = sim.run_with_config(config=cfg)

    # save to current directory. Should be an optional argument for this
    mmm_input_df.to_csv(os.path.join(output_path, "mmm_input_df.csv"), index=False)
    pd.DataFrame.from_dict(channel_roi, orient="index", columns=["true_roi"]).to_csv(
        os.path.join(output_path, "channel_roi.csv")
    )


def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "-i",
        "--input-config",
        action="store",
        help="Provides configuration file path for simulation",
    )
    arg_parser.add_argument(
        "-o",
        "--output_path",
        action="store",
        help="Provides output destination",
        default=".",
    )
    args = arg_parser.parse_args()
    run_with_config(config_path=args.input_config, output_path=args.output_path)



from pysimmmulator import load_parameters, multisimmm


def test_initiate_msim():
    msim = multisimmm()
    assert 1 == 1


def test_multiple_runs():
    cfg = load_parameters.load_config(config_path="./example_config.yaml")
    msim = multisimmm()
    msim.run(config=cfg, runs=10)
    assert len(msim.final_frames) == 10



from pysimmmulator import simmm, command_line
import subprocess


def test_cli():
    subprocess.run(["pysimmm", "-i", "example_config.yaml", "-o", "."])
    import os

    # cleanup
    os.remove("./mmm_input_df.csv")
    os.remove("./channel_roi.csv")
    assert 1 == 1



from pysimmmulator.param_handlers import (
    basic_parameters,
    baseline_parameters,
    ad_spend_parameters,
    media_parameters,
    cvr_parameters,
    adstock_parameters,
    output_parameters,
)

from .visualize import visualize

import numpy as np
import pandas as pd

import logging
logger = logging.getLogger(__name__)


class simmm(visualize):
    """Takes input of basic params and provies either piece meal or single shot
    creation of MMM data using a config file,"""

    def __init__(self, basic_params: basic_parameters = None, random_seed = None):
        self.basic_params = basic_params
        self.rng = self._create_random_factory(seed=random_seed)
        super().__init__()

    def _create_random_factory(self, seed: int) -> np.random.Generator:
        rng = np.random.default_rng(seed=seed)
        return rng
    
    def _report_random_state(self):
        return self.rng.bit_generator

    def simulate_baseline(
        self,
        base_p,
        trend_p: int,
        temp_var: int,
        temp_coef_mean: int,
        temp_coef_sd: int,
        error_std: int,
    ) -> None:
        self.baseline_params = baseline_parameters(
            basic_params=self.basic_params,
            base_p=base_p,
            trend_p=trend_p,
            temp_var=temp_var,
            temp_coef_mean=temp_coef_mean,
            temp_coef_sd=temp_coef_sd,
            error_std=error_std,
        )

        # Number of days to generate data for
        days = np.arange(0, self.basic_params.years * 365)
        # Base sales of base_p units
        base = (
            np.zeros(shape=self.basic_params.years * 365) + self.baseline_params.base_p
        )
        # Trend of trend_p extra units per day
        trend_cal = (
            self.baseline_params.trend_p / (self.basic_params.years * 365)
        ) * self.baseline_params.base_p
        trend = trend_cal * days
        # Temperature generated by a sin function and we can manipulate how much the sin function goes up or down with temp_var
        temp = self.baseline_params.temp_var * np.sin(days * 3.14 / 182.5)
        # coefficient of temperature's effect on sales will be a random variable with normal distribution
        seasonality = (
            self.rng.normal(loc=self.baseline_params.temp_coef_mean, scale=self.baseline_params.temp_coef_sd, size=1)
            * temp
        )
        # add some noise to the trend
        error = self.rng.normal(loc=0, scale=self.baseline_params.error_std, size=self.basic_params.years * 365)
        # Generate series for baseline sales
        baseline_sales = base + trend + seasonality + error
        # if error term makes baseline_sales negative, make it 0
        baseline_sales = np.where(baseline_sales < 0, 0, baseline_sales)

        self.baseline_sales_df = pd.DataFrame(
            {
                "days": days,
                "baseline_sales": baseline_sales,
                "base": base,
                "trend": trend,
                "temp": temp,
                "seasonality": seasonality,
                "error": error,
            }
        )

    def simulate_ad_spend(
        self,
        campaign_spend_mean: int,
        campaign_spend_std: int,
        max_min_proportion_on_each_channel: dict,
    ) -> None:
        ad_spend_params = ad_spend_parameters(
            campaign_spend_mean=campaign_spend_mean,
            campaign_spend_std=campaign_spend_std,
            max_min_proportion_on_each_channel=max_min_proportion_on_each_channel,
        )

        campaign_count = int(
            self.basic_params.years * 365 / self.basic_params.frequency_of_campaigns
        )

        # specify amount spent on each campaign according to a normal distribution
        campaign_spends = np.round(
            self.rng.normal(
                loc=ad_spend_params.campaign_spend_mean,
                scale=ad_spend_params.campaign_spend_std,
                size=campaign_count,
            ),
            2,
        )
        # if campaign spend number is negative, automatically make it 0
        campaign_spends[campaign_spends < 0] = 0
        campaign_channel_spend_proportions = {}
        for (
            channel,
            proportions,
        ) in ad_spend_params.max_min_proportion_on_each_channel.items():
            campaign_channel_spend_proportions[channel] = self.rng.uniform(
                low=proportions["min"],
                high=proportions["max"],
                size=campaign_count,
            )

        spend_df = pd.DataFrame(
            {
                "campaign_id": np.arange(campaign_count),
                "total_campaign_spend": campaign_spends,
            }
        )

        for channel in max_min_proportion_on_each_channel.keys():
            spend_df[channel] = np.round(
                campaign_spends * campaign_channel_spend_proportions[channel], 2
            )

        self.spend_df = spend_df.melt(
            id_vars=["campaign_id", "total_campaign_spend"],
            value_vars=self.basic_params.all_channels,
            var_name="channel",
            value_name="spend_channel",
        )
        logger.info("You have completed running step 2: Simulating ad spend.")

    def _negative_check(self, df: pd.DataFrame, column: str, channel: str) -> None:
        if df[column].min() < 0:
            sub_zero_count = (df[column] < 0).sum()
            logger.warning(
                f"There are {sub_zero_count} negative values for {channel} in {column.split('_')[1]}. Consider adjusting your distribution parameters. For now those values will be set to 0"
            )

    def _negative_replace(self, df: pd.DataFrame, column: str) -> pd.DataFrame:
        df.loc[df[column] < 0, column] = 0
        return df

    def simulate_media(
        self, true_cpm: dict, true_cpc: dict, noisy_cpm_cpc: dict
    ) -> None:
        media_params = media_parameters(true_cpm, true_cpc, noisy_cpm_cpc)
        media_params.check(basic_params=self.basic_params)

        for channel in media_params.noise_channels:
            channel_idx = self.spend_df[self.spend_df["channel"] == channel].index

            channel_noise = self.rng.normal(
                size=len(channel_idx), **noisy_cpm_cpc[channel]
            )

            channel_true_cpm_value = (
                true_cpm[channel] if channel in true_cpm.keys() else np.nan
            )
            channel_noisy_cpm_value = (
                true_cpm[channel] + channel_noise
                if channel in true_cpm.keys()
                else np.nan
            )
            self.spend_df.loc[channel_idx, "true_cpm"] = channel_true_cpm_value
            self.spend_df.loc[channel_idx, "noisy_cpm"] = channel_noisy_cpm_value

            channel_true_cpc_value = (
                true_cpc[channel] if channel in true_cpc.keys() else np.nan
            )
            channel_noisy_cpc_value = (
                true_cpc[channel] + channel_noise
                if channel in true_cpc.keys()
                else np.nan
            )
            self.spend_df.loc[channel_idx, "true_cpc"] = channel_true_cpc_value
            self.spend_df.loc[channel_idx, "noisy_cpc"] = channel_noisy_cpc_value

            self._negative_check(
                self.spend_df.loc[channel_idx], column="noisy_cpm", channel=channel
            )
            self._negative_check(
                self.spend_df.loc[channel_idx], column="noisy_cpc", channel=channel
            )

        self.spend_df = self._negative_replace(df=self.spend_df, column="noisy_cpm")
        self.spend_df = self._negative_replace(df=self.spend_df, column="noisy_cpc")

        self.spend_df["lifetime_impressions"] = np.round(
            self.spend_df["spend_channel"] / self.spend_df["noisy_cpm"] * 1000, 0
        )
        self.spend_df["lifetime_clicks"] = np.round(
            self.spend_df["spend_channel"] / self.spend_df["noisy_cpc"], 0
        )

        self.spend_df["daily_spend"] = np.round(
            self.spend_df["spend_channel"] / self.basic_params.frequency_of_campaigns, 2
        )
        self.spend_df["daily_impressions"] = np.round(
            self.spend_df["lifetime_impressions"]
            / self.basic_params.frequency_of_campaigns,
            0,
        )
        self.spend_df["daily_clicks"] = np.round(
            self.spend_df["lifetime_clicks"] / self.basic_params.frequency_of_campaigns,
            0,
        )

        logger.info("You have completed running step 3: Simulating media.")

    def simulate_cvr(self, noisy_cvr: dict) -> None:
        cvr_params = cvr_parameters(noisy_cvr)
        cvr_params.check(basic_params=self.basic_params)

        for channel in cvr_params.noise_channels:
            channel_idx = self.spend_df[self.spend_df["channel"] == channel].index

            channel_noise = self.rng.normal(size=len(channel_idx), **noisy_cvr[channel])
            self.spend_df.loc[channel_idx, "noisy_cvr"] = (
                channel_noise + self.basic_params.true_cvr[channel]
            )

            self._negative_check(
                self.spend_df.loc[channel_idx], column="noisy_cvr", channel=channel
            )
        self.spend_df = self._negative_replace(df=self.spend_df, column="noisy_cvr")
        # Daily CVR == campaign CVR, no reason to duplicate
        logger.info("You have completed running step 4: Simulating CVR.")

    def _reformat_for_mmm(self) -> None:
        date_backbone = pd.date_range(
            start=self.basic_params.start_date, end=self.basic_params.end_date, freq="D"
        )
        campaigns_in_period = (
            date_backbone.shape[0] / self.basic_params.frequency_of_campaigns
        )
        campaign_id_to_date_map = np.trunc(
            np.linspace(
                start=0, stop=campaigns_in_period - 1, num=date_backbone.shape[0]
            )
        ).astype(int)
        self.mmm_df = pd.DataFrame(
            {"date": date_backbone, "id_map": campaign_id_to_date_map}
        )
        self.mmm_df.set_index("id_map", inplace=True)

        agg_media_df = self.spend_df.groupby(["channel", "campaign_id"]).sum()[
            ["daily_impressions", "daily_clicks", "daily_spend", "noisy_cvr"]
        ]
        agg_media_df = agg_media_df.unstack(level=0)
        joined_columns = []
        for _metric, _channel in agg_media_df.columns:
            # we'll just name everything channel_metric from here. No need for daily/lifetime
            col_name = f"{_channel}_{_metric.split('_')[1]}"
            joined_columns.append(col_name)
        agg_media_df.columns = joined_columns

        self.mmm_df = self.mmm_df.join(agg_media_df)

        logger.info(
            "You have completed running step 5a: pivoting the data frame to an MMM format."
        )

    @staticmethod
    def _build_decay_vector(original_vector: pd.Series, decay_value: float) -> pd.Series:
        decayed_vector = [original_vector.values[0]]
        for i, orig_value in enumerate(original_vector.values[1:]):
            decayed_vector.append(orig_value + decay_value * decayed_vector[i])
        return pd.Series(decayed_vector)

    def _simulate_decay(self, true_lambda_decay: dict) -> None:
        for channel in true_lambda_decay.keys():
            metric = (
                "impressions"
                if channel in self.basic_params.channels_impressions
                else "clicks"
            )
            self.mmm_df[f"{channel}_{metric}_adstocked"] = self._build_decay_vector(
                original_vector=self.mmm_df[f"{channel}_{metric}"],
                decay_value=true_lambda_decay[channel],
            )

        logger.info("You have completed running step 5b: applying adstock decay.")
        # Knew I could find a better way, even better now

    def _simulate_diminishing_returns(
        self,
        alpha_saturation: dict,
        gamma_saturation: dict,
        #   , x_marginal: int = None
    ) -> None:
        for channel in alpha_saturation.keys():
            metric = (
                "impressions"
                if channel in self.basic_params.channels_impressions
                else "clicks"
            )
            target = self.mmm_df[f"{channel}_{metric}_adstocked"]
            gamma_trans = np.round(
                np.quantile(
                    np.linspace(min(target), max(target), num=100),
                    gamma_saturation[channel],
                ),
                4,
            )
            x_scurve = target ** alpha_saturation[channel] / (
                target ** alpha_saturation[channel]
                + gamma_trans ** alpha_saturation[channel]
            )
            self.mmm_df[f"{channel}_{metric}_adstocked_decay_diminishing"] = (
                x_scurve * target
            )

        logger.info(
            "You have completed running step 5c: apply diminishing marginal returns."
        )

    def simulate_decay_returns(
        self, true_lambda_decay: dict, alpha_saturation: dict, gamma_saturation: dict
    ) -> None:
        adstock_params = adstock_parameters(
            true_lambda_decay, alpha_saturation, gamma_saturation
        )
        self._reformat_for_mmm()
        self._simulate_decay(adstock_params.true_lambda_decay)
        self._simulate_diminishing_returns(
            alpha_saturation=adstock_params.alpha_saturation,
            gamma_saturation=adstock_params.gamma_saturation,
        )

        logger.info("You have completed running step 5: Simulating adstock.")

    def calculate_conversions(self):
        for channel in self.basic_params.all_channels:
            metric = (
                "impressions"
                if channel in self.basic_params.channels_impressions
                else "clicks"
            )
            self.mmm_df[f"{channel}_conversions"] = (
                self.mmm_df[f"{channel}_{metric}_adstocked_decay_diminishing"]
                * self.mmm_df[f"{channel}_cvr"]
            )

        logger.info(
            "You have completed running step 6: Calculating the number of conversions."
        )

    def consolidate_dataframe(self):
        metric_cols = []
        [metric_cols.append(f"{channel}_impressions") for channel in self.basic_params.channels_impressions]
        [metric_cols.append(f"{channel}_clicks") for channel in self.basic_params.channels_clicks]
        spend_cols = []
        [spend_cols.append(f"{channel}_spend") for channel in self.basic_params.all_channels]
        conv_cols = []
        [conv_cols.append(f"{channel}_conversions") for channel in self.basic_params.all_channels]
        self.mmm_df = self.mmm_df[["date"] + metric_cols + spend_cols + conv_cols]
        self.mmm_df["total_conversions_from_ads"] = self.mmm_df[conv_cols].sum(axis=1)
        self.mmm_df["total_revenue_from_ads"] = (
            self.mmm_df["total_conversions_from_ads"]
            * self.basic_params.revenue_per_conv
        )
        self.mmm_df["baseline_revenue"] = (
            round(self.baseline_sales_df["baseline_sales"])
            * self.basic_params.revenue_per_conv
        )
        self.mmm_df["total_revenue"] = self.mmm_df[
            ["total_revenue_from_ads", "baseline_revenue"]
        ].sum(axis=1)

        logger.info(
            "You have completed running step 7: Expanding to maximum data frame."
        )

    def calculate_channel_roi(self) -> None:
        self.channel_roi = {}
        for channel in self.basic_params.all_channels:
            total_cpa = (
                self.mmm_df[f"{channel}_spend"].sum()
                / self.mmm_df[f"{channel}_conversions"].sum()
            )
            total_roi = (self.basic_params.revenue_per_conv - total_cpa) / total_cpa
            self.channel_roi[channel] = total_roi

    def finalize_output(self, aggregation_level: str) -> None:
        output_params = output_parameters(aggregation_level)
        metric_cols = []
        [metric_cols.append(f"{channel}_impressions") for channel in self.basic_params.channels_impressions]
        [metric_cols.append(f"{channel}_clicks") for channel in self.basic_params.channels_clicks]
        spend_cols = []
        [spend_cols.append(f"{channel}_spend") for channel in self.basic_params.all_channels]

        if output_params.aggregation_level == "daily":
            self.mmm_df.set_index("date", inplace=True)
            self.final_df = self.mmm_df[metric_cols + spend_cols + ["total_revenue"]]
        else:
            self.mmm_df["week_start"] = self.mmm_df["date"] - pd.to_timedelta(
                self.mmm_df["date"].apply(lambda x: x.weekday()), unit="d"
            )
            self.final_df = (
                self.mmm_df[
                    metric_cols + spend_cols + ["total_revenue"] + ["week_start"]
                ]
                .groupby(["week_start"])
                .sum()
            )

        logger.info(
            f"You have completed running step 9: Finalization of output dataframe at the {aggregation_level} level"
        )

    def run_with_config(self, config: dict) -> set[pd.DataFrame, dict]:
        # import pysimmmulator.load_parameters as load_params
        if self.basic_params is None:
            self.basic_params = basic_parameters(**config["basic_params"])
        self.simulate_baseline(**config["baseline_params"])
        self.simulate_ad_spend(**config["ad_spend_params"])
        self.simulate_media(**config["media_params"])
        self.simulate_cvr(**config["cvr_params"])
        self.simulate_decay_returns(**config["adstock_params"])
        self.calculate_conversions()
        self.consolidate_dataframe()
        self.calculate_channel_roi()
        self.finalize_output(**config["output_params"])

        return (self.final_df, self.channel_roi)


class multisimmm(simmm):
    def __init__(self):
        super(multisimmm, self).__init__()
        self.final_frames = []
        self.rois = []

    def store_outputs(self, final_df: pd.DataFrame, channel_roi: dict):
        self.final_frames.append(final_df)
        self.rois.append(channel_roi)

    def run(self, config: dict, runs: int) -> None:
        for run in range(runs):
            frame, roi = self.run_with_config(config=config)
            self.store_outputs(final_df=frame, channel_roi=roi)
        logger.info(f"{runs} runs complete and stored")



from setuptools import setup
import os

with open(os.path.join('src/pysimmmulator', 'VERSION')) as version_file:
    version = version_file.read().strip()

setup(
    version=version,
    entry_points = {
        'console_scripts': ['pysimmm=pysimmmulator.command_line:main'],
    }
)



import pandas as pd
import matplotlib.pyplot as plt


class visualize:
    def __init__(self):
        self._viz_available = True
        self._valid_agg_levels = ['daily', 'weekly', 'monthly', 'yearly']

    def plot_spend(self, agg: str = None):
        """Plot simulated spend data based on a passed date-wise aggregation
        
        Args:
            agg (str): pick from [{', '.join(self._valid_agg_levels)}] to aggregate simulated data by"""
        assert agg in self._valid_agg_levels, f"Please select [{', '.join(self._valid_agg_levels)}] for your aggregation level. {agg} is an invalid selection."

        # prepare frame and filter columns for plotting
        self._plot_frame_overhead(agg_level=agg)
        plot_cols = self._filter_columns(columns = self.plot_frame.columns.tolist(), filter_string = '_spend')

        return self._plot_majors(columns = plot_cols)

    def plot_impressions(self, agg: str = None):
        """Plot simulated impressions data based on a passed date-wise aggregation
        
        Args:
            agg (str): pick from [{', '.join(self._valid_agg_levels)}] to aggregate simulated data by"""
        assert agg in self._valid_agg_levels, f"Please select [{', '.join(self._valid_agg_levels)}] for your aggregation level. {agg} is an invalid selection."

        # prepare frame and filter columns for plotting
        self._plot_frame_overhead(agg_level=agg)
        plot_cols = self._filter_columns(columns = self.plot_frame.columns.tolist(), filter_string = '_impressions')

        return self._plot_majors(columns = plot_cols)

    def plot_clicks(self, agg: str = None):
        """Plot simulated clicks data based on a passed date-wise aggregation
        
        Args:
            agg (str): pick from [{', '.join(self._valid_agg_levels)}] to aggregate simulated data by"""
        assert agg in self._valid_agg_levels, f"Please select [{', '.join(self._valid_agg_levels)}] for your aggregation level. {agg} is an invalid selection."

        # prepare frame and filter columns for plotting
        self._plot_frame_overhead(agg_level=agg)
        plot_cols = self._filter_columns(columns = self.plot_frame.columns.tolist(), filter_string = '_clicks')

        return self._plot_majors(columns = plot_cols)
    
    def plot_revenue(self, agg: str = None):
        """Plot simulated revenue data based on a passed date-wise aggregation
        
        Args:
            agg (str): pick from [{', '.join(self._valid_agg_levels)}] to aggregate simulated data by"""
        assert agg in self._valid_agg_levels, f"Please select [{', '.join(self._valid_agg_levels)}] for your aggregation level. {agg} is an invalid selection."

        # prepare frame and filter columns for plotting
        self._plot_frame_overhead(agg_level=agg)
        plot_cols = self._filter_columns(columns = self.plot_frame.columns.tolist(), filter_string = 'total_revenue')

        return self._plot_majors(columns = plot_cols)

    def _filter_columns(self, columns: list, filter_string: str) -> list:
        filtered_cols = []
        [filtered_cols.append(col) for col in columns if filter_string in col]
        return filtered_cols
    
    def _plot_frame_overhead(self, agg_level: int = None) -> pd.DataFrame:
        if agg_level is not None:
            self.plot_frame = self.final_df.copy()
            self.plot_frame.reset_index(inplace=True)
            self._aggregator(agg_level)
        else:
            self.plot_frame = self.final_df.copy()
    
    def _aggregator(self, agg_level: str):
        if agg_level == 'daily':
            self.plot_frame = self.plot_frame.groupby("date").sum()

        elif agg_level == 'weekly':
            self.plot_frame["week_start"] = self.plot_frame["date"] - pd.to_timedelta(
                self.plot_frame["date"].apply(lambda x: x.weekday()), unit="d"
            )
            del self.plot_frame["date"]
            self.plot_frame = self.plot_frame.groupby("week_start").sum()
            
        elif agg_level == 'monthly':
            self.plot_frame["month_start"] = self.plot_frame["date"] - pd.to_timedelta(
                self.plot_frame["date"].apply(lambda x: x.day), unit="d"
            )
            del self.plot_frame["date"]
            self.plot_frame = self.plot_frame.groupby("month_start").sum()
        
        elif agg_level == 'yearly':
            self.plot_frame["year_start"] = self.plot_frame["date"] - pd.to_timedelta(
                self.plot_frame["date"].apply(lambda x: x.timetuple()[7]), unit="d"
            )
            del self.plot_frame["date"]
            self.plot_frame = self.plot_frame.groupby("year_start").sum()


    def _plot_majors(self, columns):
        plot_subject = columns[-1].split('_')[1]
        plot_subject = plot_subject[0].upper() + plot_subject[1:]

        fig, ax = plt.subplots(1,1, figsize=(9,6), dpi=200)
        for col in columns:
            ax.plot(self.plot_frame.index, self.plot_frame[col], label=col.split('_')[0])
        ax.set_xlabel("Date")
        ax.set_ylabel(f"{plot_subject}")
        ax.set_title(f"{plot_subject} by Channel")
        fig.legend(loc="upper right")
        plt.savefig(f'{plot_subject}_by_channel.png')